#!/usr/bin/env python3

import copy
from collections.abc import Hashable
from typing import ItemsView, Iterator, KeysView, Mapping, ValuesView

from frozendict import frozendict
from typing_extensions import assert_type

fd = frozendict(a=1, b=2)

assert_type(frozendict(), frozendict[()])
assert_type(fd, frozendict[str, int])
assert_type(fd["a"], int)
assert_type(iter(fd), Iterator[str])
assert_type(frozendict.fromkeys("abc", 0), frozendict[str, int])
assert_type(fd.copy(), frozendict[str, int])
assert_type(fd.item(0), tuple[str, int])
assert_type(fd.item(), tuple[str, int])
assert_type(fd.key(0), str)
assert_type(fd.key(), str)
assert_type(fd.value(0), int)
assert_type(fd.value(), int)
assert_type(fd.get("a"), int | None)
assert_type(fd.items(), ItemsView[str, int])
assert_type(fd.keys(), KeysView[str])
assert_type(fd.values(), ValuesView[int])
assert_type(copy.copy(fd), frozendict[str, int])
assert_type(copy.deepcopy(fd), frozendict[str, int])
assert_type(reversed(fd), reversed)
assert_type(fd.delete("a"), frozendict[str, int])

assert_type(fd | {"c": 2}, frozendict[str, int])
assert_type(fd | {1: 2}, frozendict[str | int, int])
assert_type(fd | {"c": "c"}, frozendict[str, str | int])
assert_type(fd | {1: "c"}, frozendict[str | int, str | int])

assert_type(fd.setdefault("a", 0), frozendict[str, int])
assert_type(fd.setdefault("a", "a"), frozendict[str, str | int])
assert_type(fd.setdefault(0, 0), frozendict[str | int, int])
assert_type(fd.setdefault(0, "a"), frozendict[str | int, str | int])

assert_type(fd.set("abc", 1), frozendict[str, int])
assert_type(fd.set("abc", "abc"), frozendict[str, str | int])
assert_type(fd.set(1, 1), frozendict[str | int, int])
assert_type(fd.set(1, "abc"), frozendict[str | int, str | int])

# frozendict implements Mapping[K, V] and is covariant in V
vals: frozendict[str, Hashable] = fd
mapping: Mapping[str, Hashable] = fd



#!/usr/bin/env python3

import frozendict

assert frozendict.c_ext

from frozendict import frozendict
from uuid import uuid4
import pickle
from copy import copy, deepcopy
from collections.abc import MutableMapping
import tracemalloc
import gc
import sys


def getUuid():
    return str(uuid4())


class F(frozendict):
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)


class FMissing(frozendict):
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)
    
    def __missing__(self, key):
        return key


class Map(MutableMapping):
    def __init__(self, *args, **kwargs):
        self._dict = dict(*args, **kwargs)
    
    def __getitem__(self, key):
        return self._dict[key]
    
    def __setitem__(self, key, value):
        self._dict[key] = value
    
    def __delitem__(self, key):
        del self._dict[key]
    
    def __iter__(self):
        return iter(self._dict)
    
    def __len__(self):
        return len(self._dict)


def print_info(klass, iterations, func):
    try:
        name = func.__name__
    except AttributeError:
        name = func
    
    print(
        (
            f"Class = {klass.__name__} - Loops: {iterations} - " +
            f"Evaluating: {name}"
        ),
        flush=True
    )


def print_sep():
    print("-" * 72, flush=True)


def trace(iterations = 100, mult = 10.0):
    def decorator(func):
        def wrapper():
            header = (
                f"Loops: {iterations} - Multiplier: {mult} " +
                f"Evaluating: {func.__name__}"
            )
            
            print(header,  flush=True)
            
            tracemalloc.start()

            snapshot1 = tracemalloc.take_snapshot().filter_traces(
                (tracemalloc.Filter(True, __file__), )
            )

            for i in range(iterations):
                func()
            
            gc.collect()

            snapshot2 = tracemalloc.take_snapshot().filter_traces(
                (tracemalloc.Filter(True, __file__), )
            )

            top_stats = snapshot2.compare_to(snapshot1, 'lineno')
            tracemalloc.stop()
            
            for stat in top_stats:
                if stat.count_diff * mult > iterations:
                    header = (
                        f"Error. count diff: {stat.count_diff}, " +
                        f"stat: {stat}"
                    )
                    
                    print(header, flush=True)
                    
                    sys.exit(1)
        
        return wrapper
         
    return decorator


key_in = getUuid()
dict_1 = {key_in: 0}
dict_tmp = {getUuid(): i for i in range(1, 1000)}
dict_1.update(dict_tmp)
generator_1 = ((key, val) for key, val in dict_1.items())
key_notin = getUuid()
dict_2 = {getUuid(): i for i in range(1000)}
dict_3 = {i: i for i in range(1000)}
dict_hole = dict_1.copy()
del dict_hole[key_in]
dict_unashable = dict_1.copy()
dict_unashable.update({getUuid(): []})
dict_1_items = tuple(dict_1.items())
dict_1_keys = tuple(dict_1.keys())
dict_1_keys_set = set(dict_1_keys)

functions = []


@trace(iterations = 300, mult = 1.5)
def func_1():
    pickle.loads(pickle.dumps(fd_1))


functions.append(func_1)


@trace(iterations = 200, mult = 1.5)
def func_2():
    pickle.loads(pickle.dumps(iter(fd_1.keys())))


functions.append(func_2)


@trace(iterations = 300, mult = 1.5)
def func_3():
    pickle.loads(pickle.dumps(iter(fd_1.items())))


functions.append(func_3)


@trace(iterations = 400, mult = 1.5)
def func_4():
    pickle.loads(pickle.dumps(iter(fd_1.values())))


functions.append(func_4)


@trace()
def func_5():
    frozendict_class({})


functions.append(func_5)


@trace()
def func_7():
    frozendict_class([])


functions.append(func_7)


@trace()
def func_8():
    frozendict_class({}, **{})


functions.append(func_8)


@trace()
def func_9():
    frozendict_class(**dict_1)


functions.append(func_9)


@trace()
def func_10():
    frozendict_class(dict_1, **dict_2)


functions.append(func_10)


@trace()
def func_11():
    frozendict_class(fd_1)


functions.append(func_11)


@trace()
def func_12():
    frozendict_class(generator_1)


functions.append(func_12)


@trace()
def func_13():
    frozendict_class(dict_hole)


functions.append(func_13)


@trace()
def func_14():
    fd_1.copy()


functions.append(func_14)


@trace()
def func_15():
    fd_1 == dict_1


functions.append(func_15)


@trace()
def func_16():
    fd_1 == fd_1


functions.append(func_16)


@trace()
def func_17():
    fd_1 != dict_hole


functions.append(func_17)


@trace()
def func_18():
    fd_1 != dict_2


functions.append(func_18)


@trace()
def func_19():
    fd_1 == dict_hole


functions.append(func_19)


@trace()
def func_20():
    fd_1 == dict_2


functions.append(func_20)


@trace()
def func_21():
    frozendict_class(dict_1)


functions.append(func_21)


@trace()
def func_22():
    frozendict_class(dict_1_items)


functions.append(func_22)


@trace()
def func_23():
    tuple(fd_1.keys())


functions.append(func_23)


@trace()
def func_24():
    tuple(fd_1.values())


functions.append(func_24)


@trace()
def func_25():
    tuple(fd_1.items())


functions.append(func_25)


@trace()
def func_26():
    frozendict_class.fromkeys(dict_1)


functions.append(func_26)


@trace()
def func_27():
    frozendict_class.fromkeys(dict_1, 1)


functions.append(func_27)


@trace()
def func_28():
    frozendict_class.fromkeys(dict_1_keys)


functions.append(func_28)


@trace()
def func_29():
    frozendict_class.fromkeys(dict_1_keys, 1)


functions.append(func_29)


@trace()
def func_30():
    frozendict_class.fromkeys(dict_1_keys_set)


functions.append(func_30)


@trace()
def func_31():
    frozendict_class.fromkeys(dict_1_keys_set, 1)


functions.append(func_31)


@trace()
def func_32():
    repr(fd_1)


functions.append(func_32)


@trace()
def func_33():
    fd_1 | dict_2


functions.append(func_33)


@trace()
def func_34():
    hash(fd_1)


functions.append(func_34)


@trace()
def func_35():
    frozendict_class() == frozendict_class()


functions.append(func_35)


@trace()
def func_36():
    tuple(reversed(fd_1))


functions.append(func_36)


@trace()
def func_37():
    tuple(reversed(fd_1.keys()))


functions.append(func_37)


@trace()
def func_38():
    tuple(reversed(fd_1.items()))


functions.append(func_38)


@trace()
def func_39():
    tuple(reversed(fd_1.values()))


functions.append(func_39)


@trace()
def func_40():
    iter(fd_1).__length_hint__()
    

functions.append(func_40)


@trace()
def func_41():
    len(fd_1)


functions.append(func_41)


@trace()
def func_42():
    len(fd_1.keys())


functions.append(func_42)


@trace()
def func_43():
    len(fd_1.items())


functions.append(func_43)


@trace()
def func_44():
    len(fd_1.values())


functions.append(func_44)


@trace()
def func_45():
    fd_1.keys().mapping == fd_1


functions.append(func_45)


@trace()
def func_46():
    fd_1.items().mapping == fd_1


functions.append(func_46)


@trace()
def func_47():
    fd_1.values().mapping == fd_1


functions.append(func_47)


@trace()
def func_48():
    fd_1[key_in]


functions.append(func_48)


@trace()
def func_49():
    fd_1.get(key_in)


functions.append(func_49)


@trace()
def func_50():
    fd_1.get(key_notin)


functions.append(func_50)


@trace()
def func_51():
    fd_1.get(key_notin, 1)


functions.append(func_51)


@trace()
def func_52():
    key_in in fd_1


functions.append(func_52)


@trace()
def func_53():
    key_notin in fd_1


functions.append(func_53)


@trace()
def func_54():
    fd_1.copy()


functions.append(func_54)


@trace()
def func_55():
    copy(fd_1)


functions.append(func_55)


@trace()
def func_56():
    deepcopy(fd_1)


functions.append(func_56)


@trace()
def func_57():
    deepcopy(fd_unashable)


functions.append(func_57)


@trace()
def func_58():
    fd_1.keys() == dict_1.keys()


functions.append(func_58)


@trace()
def func_59():
    fd_1.items() == dict_1.items()


functions.append(func_59)


@trace()
def func_60():
    key_notin in fd_1.keys()


functions.append(func_60)


@trace()
def func_61():
    (key_notin, 0) in fd_1.items()


functions.append(func_61)


@trace()
def func_62():
    FMissing(fd_1)[0]


functions.append(func_62)


@trace()
def func_63():
    mp = Map(dict_1)
    frozendict_class(mp) == dict_1


functions.append(func_63)


@trace()
def func_64():
    fd_1.keys().isdisjoint(dict_3)


functions.append(func_64)


@trace()
def func_65():
    fd_1.keys().isdisjoint(fd_1)


functions.append(func_65)


@trace()
def func_66():
    fd_1.items().isdisjoint(dict_3.items())


functions.append(func_66)


@trace()
def func_67():
    fd_1.items().isdisjoint(fd_1.items())


functions.append(func_67)


@trace()
def func_68():
    fd_unashable.keys() - fd_1.keys()


functions.append(func_68)


@trace()
def func_69():
    fd_1.items() - frozendict_class(dict_hole).items()


functions.append(func_69)


@trace()
def func_70():
    fd_1.keys() & frozendict_class(dict_hole).keys()


functions.append(func_70)


@trace()
def func_71():
    fd_1.items() & frozendict_class(dict_hole).items()


functions.append(func_71)


@trace()
def func_72():
    fd_1.keys() | frozendict_class(dict_2).keys()


functions.append(func_72)


@trace()
def func_73():
    fd_1.items() | frozendict_class(dict_2).items()


functions.append(func_73)


@trace()
def func_74():
    fd_1.keys() ^ frozendict_class(dict_hole).keys()


functions.append(func_74)


@trace()
def func_75():
    fd_1.items() ^ frozendict_class(dict_hole).items()


functions.append(func_75)


@trace()
def func_76():
    frozendict_class(dict_unashable)


functions.append(func_76)


@trace()
def func_77():
    frozendict_class(dict_3)


functions.append(func_77)


@trace()
def func_78():
    frozendict_class()


functions.append(func_78)


@trace()
def func_79():
    frozendict_class(dict_hole).keys() < fd_1.keys()


functions.append(func_79)


@trace()
def func_80():
    frozendict_class(dict_hole).keys() <= fd_1.keys()


functions.append(func_80)


@trace()
def func_81():
    frozendict_class(dict_hole).items() < fd_1.items()


functions.append(func_81)


@trace()
def func_82():
    fd_1.keys() > frozendict_class(dict_hole).keys()


functions.append(func_82)


@trace()
def func_83():
    fd_1.keys() >= frozendict_class(dict_hole).keys()


functions.append(func_83)


@trace()
def func_84():
    fd_1.items() > frozendict_class(dict_hole).items()


functions.append(func_84)


@trace()
def func_85():
    fd_1.items() >= frozendict_class(dict_hole).items()


functions.append(func_85)


@trace()
def func_86():
    fd_1.set(key_in, 1000)


functions.append(func_86)


@trace()
def func_87():
    fd_1.set(key_notin, 1000)


functions.append(func_87)


@trace()
def func_88():
    fd_1.delete(key_in)


functions.append(func_88)


@trace()
def func_89():
    fd_1.setdefault(key_in)


functions.append(func_89)


@trace()
def func_90():
    fd_1.setdefault(key_notin)


functions.append(func_90)


@trace()
def func_91():
    fd_1.setdefault(key_notin, 1000)


functions.append(func_91)


@trace()
def func_92():
    fd_1.key()


functions.append(func_92)


@trace()
def func_93():
    fd_1.key(0)


functions.append(func_93)


@trace()
def func_94():
    fd_1.key(-1)


functions.append(func_94)


@trace()
def func_95():
    fd_1.value()


functions.append(func_95)


@trace()
def func_96():
    fd_1.value(0)


functions.append(func_96)


@trace()
def func_97():
    fd_1.value(-1)


functions.append(func_97)


@trace()
def func_98():
    fd_1.item()


functions.append(func_98)


@trace()
def func_99():
    fd_1.item(0)


functions.append(func_99)


@trace()
def func_100():
    fd_1.item(-1)


functions.append(func_100)


@trace()
def func_101():
    for _ in fd_1:
        pass


functions.append(func_101)


@trace()
def func_102():
    for _ in iter(fd_1):
        pass


functions.append(func_102)


@trace()
def func_103():
    for _ in fd_1.keys():
        pass


functions.append(func_103)


@trace()
def func_104():
    for _ in fd_1.values():
        pass


functions.append(func_104)


@trace()
def func_105():
    for _ in fd_1.items():
        pass


functions.append(func_105)


@trace()
def func_106():
    try:
        hash(fd_unashable)
    except TypeError:
        pass


functions.append(func_106)


@trace()
def func_107():
    try:
        fd_1[key_notin]
    except KeyError:
        pass


functions.append(func_107)


@trace()
def func_108():
    try:
        fd_1.key(len(fd_1))
    except IndexError:
        pass
    

functions.append(func_108)


@trace()
def func_109():
    try:
        fd_1.value(len(fd_1))
    except IndexError:
        pass


functions.append(func_109)


@trace()
def func_110():
    try:
        fd_1.item(len(fd_1))
    except IndexError:
        pass


functions.append(func_110)


print_sep()

for frozendict_class in (frozendict, F):
    print_info(
        frozendict_class, 
        1, 
        "frozendict_class(dict_1)"
    )
    
    fd_1 = frozendict_class(dict_1)
    print_sep()
    print_info(
        frozendict_class, 
        1, 
        "frozendict_class(dict_unashable)"
    )
    
    fd_unashable = frozendict_class(dict_unashable)
    print_sep()
    
    for function in functions:
        print(
            f"Class = {frozendict_class.__name__} - ", 
            flush = True,
            end = ""
        )
        
        function()
        
        print_sep()



from copy import copy, deepcopy
from .base import FrozendictTestBase


class FrozendictSubclassOnlyTest(FrozendictTestBase):
    _FrozendictMissingClass = None
    
    @property
    def FrozendictMissingClass(self):
        val = self._FrozendictMissingClass
        
        if val is None:
            raise ValueError("FrozendictMissingClass is None")
        
        return val
    
    @FrozendictMissingClass.setter
    def FrozendictMissingClass(self, val):
        self._FrozendictMissingClass = val
    
    ####################################################################
    # tests
    
    def test_empty_sub(self, fd_empty):
        fd_empty_2 = self.FrozendictClass({})
        fd_empty_3 = self.FrozendictClass([])
        fd_empty_4 = self.FrozendictClass({}, **{})
        
        assert fd_empty == fd_empty_2 == fd_empty_3 == fd_empty_4
        assert fd_empty is not fd_empty_2 is not fd_empty_3
        assert fd_empty is not fd_empty_4
    
    def test_constructor_self_sub(self, fd):
        fd_clone = self.FrozendictClass(fd)
        assert fd == fd_clone
        assert fd is not fd_clone

    def test_copy_sub(self, fd):
        fd_copy = fd.copy()
        assert fd_copy == fd
        assert fd_copy is not fd

    def test_copycopy_sub(self, fd, fd_unhashable):
        fd_copy = copy(fd)
        fd_copy_unhashable = copy(fd_unhashable)
        assert fd_copy == fd
        assert fd_copy_unhashable == fd_unhashable
        assert fd_copy is not fd
        assert fd_copy_unhashable is not fd_unhashable

    def test_deepcopy_sub(self, fd):
        fd_copy = deepcopy(fd)
        assert fd_copy == fd
        assert fd_copy is not fd

    def test_init_sub(self, fd):
        fd_copy = fd.copy()
        fd_clone = self.FrozendictClass(dict(fd))
        fd.__init__({"Trump": "Donald"})
        assert fd_copy == fd
        assert fd_clone == fd
        assert fd_copy is not fd

    def test_del_empty_sub(self, fd_empty):
        f = self.FrozendictClass({1: 2})
        f2 = f.delete(1)
        assert f2 == fd_empty
        assert f2 is not fd_empty

    def test_missing(self, fd):
        fd_missing = self.FrozendictMissingClass(fd)
        assert fd_missing[0] == 0



import io
import pickle
import pytest
from copy import copy, deepcopy
from .base import FrozendictTestBase


class FrozendictOnlyTest(FrozendictTestBase):
    def test_empty(self, fd_empty):
        fd_empty_set = self.FrozendictClass({})
        fd_empty_list = self.FrozendictClass([])
        fd_empty_dict = self.FrozendictClass({}, **{})
        
        assert fd_empty is fd_empty_set is fd_empty_list
        assert fd_empty is fd_empty_dict

    def test_constructor_self_1(self, fd):
        assert fd is self.FrozendictClass(fd)

    def test_vars(self, fd):
        with pytest.raises(TypeError):
            vars(fd)

    def test_setattr(self, fd):
        with pytest.raises(AttributeError):
            fd._initialized = False

    def test_copy(self, fd):
        assert fd.copy() is fd

    def test_copycopy(self, fd, fd_unhashable):
        assert copy(fd) is fd
        assert copy(fd_unhashable) is fd_unhashable

    def test_deepcopy(self, fd):
        assert deepcopy(fd) is fd

    def test_init(self, fd):
        fd_copy = fd.copy()
        fd_clone = self.FrozendictClass(dict(fd))
        fd.__init__({"Trump": "Donald"})
        assert fd_copy is fd
        assert fd_clone == fd

    def test_del_empty(self):
        f = self.FrozendictClass({1: 2})
        assert f.delete(1) is self.FrozendictClass()

    def test_pickle_core(self, fd):
        class CustomUnpickler(pickle.Unpickler):
            def find_class(self, module, name):
                assert module == 'frozendict'
                assert name == 'frozendict'
                return super().find_class('frozendict.core', name)

        dump = pickle.dumps(fd)
        assert dump
        assert CustomUnpickler(io.BytesIO(dump)).load() == fd



r"""
Provides frozendict, a simple immutable dictionary.
"""

try:
    from ._frozendict import *
    c_ext = True
    del _frozendict
except ImportError:
    from ._frozendict_py import *
    c_ext = False

from .version import version as __version__
from . import monkeypatch
from .cool import *


def _getFrozendictJsonEncoder(BaseJsonEncoder = None):
    if BaseJsonEncoder is None:
        from json.encoder import JSONEncoder
        
        BaseJsonEncoder = JSONEncoder
    
    class FrozendictJsonEncoderInternal(BaseJsonEncoder):
        def default(self, obj):
            if isinstance(obj, frozendict):
                # TODO create a C serializer
                return dict(obj)
            
            return BaseJsonEncoder.default(self, obj)
    
    return FrozendictJsonEncoderInternal


FrozendictJsonEncoder = _getFrozendictJsonEncoder()
monkeypatch.patchOrUnpatchAll(patch = True, warn = False)


from collections.abc import Mapping
Mapping.register(frozendict)
del Mapping


if c_ext:
    __all__ = (frozendict.__name__, )
else:
    __all__ = _frozendict_py.__all__
    del _frozendict_py

# TODO deprecated, to remove in future versions
FrozenOrderedDict = frozendict

__all__ += cool.__all__
__all__ += (FrozendictJsonEncoder.__name__, "FrozenOrderedDict")

del cool



from copy import deepcopy


def immutable(self, *args, **kwargs):
    r"""
    Function for not implemented method since the object is immutable
    """
    
    raise AttributeError(
        f"'{self.__class__.__name__}' object is read-only"
    )


_empty_frozendict = None


class frozendict(dict):
    r"""
    A simple immutable dictionary.

    The API is the same as `dict`, without methods that can change the
    immutability. In addition, it supports __hash__().
    """
    
    __slots__ = (
        "_hash",
    )
    
    @classmethod
    def fromkeys(cls, *args, **kwargs):
        r"""
        Identical to dict.fromkeys().
        """
        
        return cls(dict.fromkeys(*args, **kwargs))
    
    def __new__(e4b37cdf_d78a_4632_bade_6f0579d8efac, *args, **kwargs):
        cls = e4b37cdf_d78a_4632_bade_6f0579d8efac
        
        has_kwargs = bool(kwargs)
        continue_creation = True
        
        # check if there's only an argument and it's of the same class
        if len(args) == 1 and not has_kwargs:
            it = args[0]
            
            # no isinstance, to avoid subclassing problems
            if it.__class__ == frozendict and cls == frozendict:
                self = it
                continue_creation = False
        
        if continue_creation:
            self = dict.__new__(cls, *args, **kwargs)
            
            dict.__init__(self, *args, **kwargs)
            
            # empty singleton - start
            
            if self.__class__ == frozendict and not len(self):
                global _empty_frozendict
                
                if _empty_frozendict is None:
                    _empty_frozendict = self
                else:
                    self = _empty_frozendict
                    continue_creation = False
            
            # empty singleton - end
            
            if continue_creation:
                object.__setattr__(self, "_hash", -1)
        
        return self
    
    def __init__(self, *args, **kwargs):
        pass
    
    def __hash__(self, *args, **kwargs):
        r"""
        Calculates the hash if all values are hashable, otherwise
        raises a TypeError.
        """
        
        if self._hash != -1:
            _hash = self._hash
        else:
            fs = frozenset(self.items())
            _hash = hash(fs)
            
            object.__setattr__(self, "_hash", _hash)
        
        return _hash
    
    def __repr__(self, *args, **kwargs):
        r"""
        Identical to dict.__repr__().
        """
        
        body = super().__repr__(*args, **kwargs)
        klass = self.__class__
        
        if klass == frozendict:
            name = f"frozendict.{klass.__name__}"
        else:
            name = klass.__name__
        
        return f"{name}({body})"
    
    def copy(self):
        r"""
        Return the object itself, as it's an immutable.
        """
        
        klass = self.__class__
        
        if klass == frozendict:
            return self
        
        return klass(self)
    
    def __copy__(self, *args, **kwargs):
        r"""
        See copy().
        """
        
        return self.copy()
    
    def __deepcopy__(self, memo, *args, **kwargs):
        r"""
        As for tuples, if hashable, see copy(); otherwise, it returns a
        deepcopy.
        """
        
        klass = self.__class__
        return_copy = klass == frozendict
        
        if return_copy:
            try:
                hash(self)
            except TypeError:
                return_copy = False
        
        if return_copy:
            return self.copy()
        
        tmp = deepcopy(dict(self))
        
        return klass(tmp)
    
    def __reduce__(self, *args, **kwargs):
        r"""
        Support for `pickle`.
        """
        
        return (self.__class__, (dict(self),))
    
    def set(self, key, val):
        new_self = deepcopy(dict(self))
        new_self[key] = val
        
        return self.__class__(new_self)
    
    def setdefault(self, key, default=None):
        if key in self:
            return self
        
        new_self = deepcopy(dict(self))
        
        new_self[key] = default
        
        return self.__class__(new_self)
    
    def delete(self, key):
        new_self = deepcopy(dict(self))
        del new_self[key]
        
        if new_self:
            return self.__class__(new_self)
        
        return self.__class__()
    
    def _get_by_index(self, collection, index):
        try:
            return collection[index]
        except IndexError:
            maxindex = len(collection) - 1
            name = self.__class__.__name__
            raise IndexError(
                f"{name} index {index} out of range {maxindex}"
            ) from None
    
    def key(self, index=0):
        collection = tuple(self.keys())
        
        return self._get_by_index(collection, index)
    
    def value(self, index=0):
        collection = tuple(self.values())
        
        return self._get_by_index(collection, index)
    
    def item(self, index=0):
        collection = tuple(self.items())
        
        return self._get_by_index(collection, index)
    
    def __setitem__(self, key, val, *args, **kwargs):
        raise TypeError(
            f"'{self.__class__.__name__}' object doesn't support item "
            "assignment"
        )
    
    def __delitem__(self, key, *args, **kwargs):
        raise TypeError(
            f"'{self.__class__.__name__}' object doesn't support item "
            "deletion"
        )


def frozendict_or(self, other, *args, **kwargs):
    res = {}
    res.update(self)
    res.update(other)
    
    return self.__class__(res)


frozendict.__or__ = frozendict_or
frozendict.__ior__ = frozendict_or

try:
    frozendict.__reversed__
except AttributeError:
    def frozendict_reversed(self, *args, **kwargs):
        return reversed(tuple(self))
    
    
    frozendict.__reversed__ = frozendict_reversed

frozendict.clear = immutable
frozendict.pop = immutable
frozendict.popitem = immutable
frozendict.update = immutable
frozendict.__delattr__ = immutable
frozendict.__setattr__ = immutable
frozendict.__module__ = 'frozendict'

__all__ = (frozendict.__name__,)
